# -*- coding: utf-8 -*-
"""Gnarwl Forwarder class file."""

import base64


class Forwarder(object):
    """Forwarder class."""

    def __init__(self, update, record):
        """Initialize a person instance."""
        self.record = record
        self.update = update
        self.verbose = update.verbose

        self.username = str(record['username'])
        self.email = '%s@broadinstitute.org' % (self.username)

        self.created = record.get('created')
        self.modified = record.get('modified')

        self.enddate = record.get('enddate')
        self.forward = str(record.get('forward', ''))
        self.sponsor = record.get('sponsor')

        if self.username == 'nagios':
            self.active = 'FALSE'
            self.message = None
            self.password = 'nagios'
            self.first_name = 'Nagios'
            self.last_name = 'Check'
            self.full_name = 'Nagios Check'
        else:
            self.active = 'TRUE'
            self.password = '*'
            self.message = ''
            raw_message = record['message']
            if raw_message:
                try:
                    self.message = base64.b64decode(raw_message).decode('utf-8')
                except Exception as error:
                    print(f"ERROR: Failed to decode Gnarwl message for: {self.username}")
                    print(error)
                    print(raw_message)
            self.get_person()

    def get_person(self):
        """Return a person record if one exists."""
        self.person = self.update.usernames.get(self.username)
        self.first_name = self.username
        self.last_name = self.username
        self.full_name = self.username
        if self.person:
            self.first_name = str(self.person['first_name'])
            self.last_name = str(self.person['last_name'])
            self.full_name = str(self.person['full_name'])

    def to_ldap(self):
        """Return a record for LDAP."""
        record = {
            'cn': [self.username],
            'displayName': [self.full_name],
            'givenName': [self.first_name],
            # 'mail': [self.email],
            'objectClass': [
                'inetOrgPerson',
                'extensibleObject',
                'top',
            ],
            'sn': [self.last_name],
            'uid': [self.username],
            # 'userPassword': [self.password],
            'vacationActive': [self.active],
        }
        # nagios user needs a password
        if self.username == 'nagios':
            record['userPassword'] = [self.password]
        # other users get mail
        else:
            record['mail'] = [self.email]

        if self.message is not None:
            record['vacationInfo'] = [self.message]

        return record



# -*- coding: utf-8 -*-
"""LDAP Update class for Gnarwl."""

from bits.ldap.server import Server


class Gnarwl(Server):
    """Gnarwl LDAP Update class."""

    def prepare_entries(self, ldap):
        """Prepare LDAP entries."""
        self.ldap = ldap
        # forwarders
        if self.verbose:
            print('  Preparing forwarders for Gnarwl LDAP...')
        self.new_entries = self.prepare_forwarders()
        if self.verbose:
            print('  Found %s Gnarwl forwarders' % (len(self.new_entries)))

    def prepare_forwarders(self):
        """Prepare Gnarwl forwarders."""
        forwarders = {}
        for record in self.update.gnarwl_records:
            dn = self.get_gnarwl_dn(record)
            forwarders[dn] = record.to_ldap()
        return forwarders



# -*- coding: utf-8 -*-
"""LDAP Update class file."""

from bits.ldap import LDAP

# ldap server types
# from bits.ldap.ad import AD
# from bits.ldap.broad import Broad
from bits.ldap.gnarwl import Gnarwl
from bits.ldap.google import Google
# from bits.ldap.localmail import Localmail
from bits.ldap.servicenow import ServiceNow

# object types
from bits.ldap.account import Account
from bits.ldap.forwarder import Forwarder
from bits.ldap.person import Person


class LDAPUpdate(object):
    """LDAPUpdate class."""

    enterprise_group = 'enterprise-gsuite-licenses@broadinstitute.org'

    def __init__(self, auth, settings):
        """Initialize a class instance."""
        self.auth = auth
        self.settings = settings
        self.verbose = auth.verbose

        # ldap server configs
        self.ldap_servers = settings.get('ldap_servers')

        self.other_accounts = {}

    def assemble_records(self, servers):
        """Assemble data into objects."""
        if self.verbose:
            print('Assembling data for LDAP update...')

        # assemble people data
        people_types = set(['ad', 'google', 'servicenow'])
        if self.types.intersection(people_types):
            self.assemble_people_records()

        # assemble account data
        if 'google' in self.types:
            self.assemble_account_records()

        # assemble gnarwl data
        if 'gnarwl' in self.types:
            self.assemble_gnarwl_records()

        if self.verbose:
            print

    def assemble_account_records(self):
        """Assemble people data."""
        if self.verbose:
            print('  Assembling Account data...')

        accounts = []

        for account in self.other_accounts:
            accounts.append(Account(self, account))

        if self.verbose:
            print('    Created %s Account objects.' % (len(accounts)))

        self.accounts_records = accounts

    def assemble_gnarwl_records(self):
        """Assemble gnarwl data."""
        if self.verbose:
            print('  Assembling Gnarwl data...')

        forwarders = [Forwarder(self, {'username': 'nagios'})]

        for username in self.gnarwl:

            # add the user as a forwarder
            record = self.gnarwl[username]
            forwarders.append(Forwarder(self, record))

            # also add nicknames as forwarders
            if username in self.nicknames:
                nicknames = self.nicknames[username]['nicknames']
                for nickname in nicknames:
                    record['username'] = nickname
                    forwarders.append(Forwarder(self, record))

        if self.verbose:
            print('    Created %s Gnarwl objects.' % (len(forwarders)))

        self.gnarwl_records = forwarders

    def assemble_people_records(self):
        """Assemble people data."""
        if self.verbose:
            print('  Assembling People data...')

        people = {}

        for pid in self.people:
            people[pid] = Person(self, self.people[pid])

        if self.verbose:
            print('    Created %s Person objects.' % (len(people)))

        self.people_records = people

    def get_data(self, servers):
        """Get data needed for the LDAP update."""
        if self.verbose:
            print('Getting data for LDAP update...')

        # determine which server types are being updated
        # Types: ad, gnarwl, google, localmail, servicenow
        self.types = self.get_server_types(servers)
        if self.verbose:
            print('  Types: %s' % (', '.join(sorted(self.types))))

        # get people data
        people_types = set(['gnarwl', 'google', 'servicenow'])
        if self.types.intersection(people_types):
            self.get_people()
            self.get_nicknames()

        # additional data
        other_types = set(['google', 'servicenow'])
        if self.types.intersection(other_types):
            self.get_ad_users()
            self.get_desks()
            self.get_enterprise_users()
            self.get_google_people()
            self.get_google_users()
            self.get_hold()
            self.get_other_accounts()
            self.get_phones()
            self.get_resources()
            self.get_shared_contacts()
            self.get_trusted_testers()

        # get gnarwl data
        if 'gnarwl' in self.types:
            self.get_gnarwl()

        # get localhost data
        # if 'localhost' in types:
        #     self.get_localhost_data()

        if self.verbose:
            print

    # def get_gnarwl(self):
    #     """Prepare gnarwl data."""

    # def get_localmail(self):
    #     """Prepare localmail data."""

    def get_ad_users(self):
        """Get AD users data."""
        ad = self.auth.ad()
        attrlist = [
            'gidNumber',
            'homeDirectory',
            'loginShell',
            'sAMAccountName',
            'uidNumber',
            'unixHomeDirectory',
        ]
        if self.verbose:
            print('  Getting users from AD...')
        ad_users = ad.getUsers(attrlist=attrlist, full=False)
        if self.verbose:
            print('    Found %s AD users.' % (len(ad_users)))
        self.ad_users = ad_users

    def get_desks(self):
        """Get desks data."""
        s = self.auth.space()
        if self.verbose:
            print('  Getting seats from Space...')
        seats = s.getSeats()
        if self.verbose:
            print('    Found %s seats.' % (len(seats)))
        desks = {}
        for seat_id in sorted(seats):
            seat = seats[seat_id]
            pid = seat['pid']
            if pid not in desks:
                desks[pid] = []
            desks[pid].append(seat)
        self.desks = desks
        # get buildings too
        self.buildings = s.getBuildings()

    def get_enterprise_users(self):
        """Get users with gsuite enterprise license override."""
        g = self.auth.google()
        g.auth_service_account(g.scopes, g.subject)
        if self.verbose:
            print('  Getting G Suite Enterprise users from Google Groups...')
        enterprise = {}
        for member in g.directory().get_derived_members(
            self.enterprise_group,
        ):
            uid = member['id']
            enterprise[uid] = member
        if self.verbose:
            print('    Found %s Enterprise users.' % (len(enterprise)))
        self.enterprise_users = enterprise

    def get_gnarwl(self):
        """Get gnarwl forwarders."""
        m = self.auth.mongo()
        self.gnarwl = m.get_collection_dict('gnarwl')

    def get_google_people(self):
        """Get google people."""
        m = self.auth.mongo()
        self.google_people = m.get_collection_dict('google_people')

    def get_google_users(self):
        """Get google users."""
        g = self.auth.google()
        g.auth_service_account(g.scopes, g.subject)
        user_fields = [
            'id',
            'primaryEmail',
        ]
        fields = 'nextPageToken,users(%s)' % (','.join(user_fields))
        if self.verbose:
            print('  Getting users from Google.')
        google_users = g.directory().get_users_dict(fields=fields, key='primaryEmail')
        if self.verbose:
            print('    Found %s Google users' % (len(google_users)))
        self.google_users = google_users

    def get_hold(self):
        """Get hold."""
        a = self.auth.accounts()
        if self.verbose:
            print('  Getting users on hold...')
        hold = a.hold
        if self.verbose:
            print('    Found %s users on hold.' % (len(hold)))
        self.hold = hold

    def get_nicknames(self):
        """Get nicknames."""
        m = self.auth.mongo()
        self.nicknames = m.get_collection_dict('nicknames')
        # also add nicknames to usernames
        for username in self.nicknames:
            if username not in self.usernames:
                continue
            person = self.usernames[username]
            nicknames = self.nicknames[username].get('nicknames', [])
            for nickname in nicknames:
                if nickname in self.usernames:
                    print('ERROR: Duplicate username: %s' % (nickname))
                    continue
                self.usernames[nickname] = person

    def get_other_accounts(self):
        """Get other_accounts."""
        m = self.auth.mongo()
        other_accounts = m.get_collection_dict('other_accounts')

        accounts = []
        for oid in other_accounts:
            account = other_accounts[oid]

            # get google usernames from record
            google_username = account.get('google_username')
            google_test_username = account.get('google_test_username')
            google_us_username = account.get('google_us_username')
            google_us_test_username = account.get('google_us_test_username')

            # skip records with no google username
            if (
                not google_username
                and not google_test_username
                and not google_us_username
                and not google_us_test_username
            ):
                continue

            accounts.append(account)
        self.other_accounts = accounts

    def get_people(self):
        """Get people data."""
        p = self.auth.people()
        if self.verbose:
            print('  Getting people from People...')
        people = p.getPeople()
        if self.verbose:
            print('    Found %s people.' % (len(people)))
        self.people = people

        # get people by username
        usernames = {}
        for pid in self.people:
            person = self.people[pid]
            username = person['username']
            if username:
                usernames[username] = person
        self.usernames = usernames

    def get_phones(self):
        """Get phone data."""
        p = self.auth.people()
        if self.verbose:
            print('  Getting phones from People...')
        phones = p.getTelephones()
        if self.verbose:
            print('    Found %s phones.' % (len(phones)))
        self.phones = phones

    def get_resources(self):
        """Get calendar resources from Google."""
        g = self.auth.google()
        g.auth_service_account(g.scopes, g.subject)
        if self.verbose:
            print('  Getting resouces from Google...')
        resources = g.directory().get_resource_calendars()
        if self.verbose:
            print('    Found %s resources.' % (len(resources)))
        self.resources = resources

    def get_server_types(self, servers):
        """Return a set of types from the servers list."""
        types = []
        for server in servers:
            config = self.ldap_servers[server]
            server_type = config['type']
            types.append(server_type)
        return set(types)

    def get_shared_contacts(self):
        """Prepare shared contacts data."""
        m = self.auth.mongo()
        self.shared_contacts = m.get_collection_dict('google_shared_contacts')

    def get_trusted_testers(self):
        """Get trusted testers."""
        m = self.auth.mongo()
        self.trusted_testers = m.get_collection_dict('trusted_testers')

    def prepare_entries(self, server_type, ldap):
        """Prepare entries for an LDAP server update."""
        # get update class for Gnarwl records
        if server_type == 'gnarwl':
            update = Gnarwl(self)

        # get update class for Google records
        if server_type == 'google':
            update = Google(self)

        # get update class for ServiceNow servers
        elif server_type == 'servicenow':
            update = ServiceNow(self)

        # prepare entries for this class
        update.prepare_entries(ldap)

        # return the update class that will be used for this server
        return update

    def update(self, servers):
        """Update a list of LDAP servers."""
        # retrieve raw data from sources
        self.get_data(servers)

        # assemble raw data into objects
        self.assemble_records(servers)

        # update ldap servers
        self.update_servers(servers)

    def update_all(self):
        """Update all LDAP servers."""
        self.update(self.ldap_servers)

    def update_server(self, server):
        """Update a single LDAP server."""
        # get ldap config for this server
        config = self.ldap_servers[server]

        # get server type
        server_type = config.get('type')

        # skip unsupported server types
        if server_type not in ['gnarwl', 'google', 'servicenow']:
            return

        print('Server: %s' % (server))

        # connect to ldap server
        ldap = LDAP(
            base_dn=config.get('base_dn'),
            bind_dn=config.get('bind_dn'),
            bind_pw=config.get('bind_pw'),
            bitsdb_key=config.get('bitsdb_key'),
            domains=config.get('domains'),
            server_type=config.get('type'),
            uri=config.get('uri'),
            verbose=self.verbose,
        )

        # prepare entries for this specific server
        update = self.prepare_entries(server_type, ldap)

        # skip servers with no update class
        if not update.new_entries:
            if self.verbose:
                print('  No new entries to compare. Skipping.\n')
            return

        # peform updates for this server
        update.perform_updates(server)

        if self.verbose:
            print

    def update_servers(self, servers):
        """Update the listed servers."""
        # update each server
        for server in sorted(servers):
            self.update_server(server)



# -*- coding: utf-8 -*-
"""Account class file."""


class Account(object):
    """Account class."""

    def __init__(self, update, account):
        """Initialize an other account instance."""
        # get basic information
        self.description = account.get('description')
        self.first_name = account.get('first_name')
        self.last_name = account.get('last_name')
        self.name = account.get('name')
        self.type = account.get('type')

        # get full name
        self.full_name = '%s %s' % (self.first_name, self.last_name)

        # get displayName
        if self.description:
            self.displayName = str(self.description.strip())
        elif self.name:
            self.displayName = str(self.name.strip())
        else:
            self.displayName = str(self.full_name.strip())

        # get names
        names = self.displayName.split(' ')
        if len(names) < 2:
            names.append('Account')
        # first/last
        self.first_name = str(names[0])
        self.last_name = str(' '.join(names[1:]))

        # google information
        self.google_license = account.get('google_license')
        self.google_ou = account.get('google_ou')

        # broadinstitute.org
        self.google_secure = account.get('google_secure')
        self.google_username = account.get('google_username')
        # test.broadinstitute.org
        self.google_test_secure = account.get('google_test_secure')
        self.google_test_username = account.get('google_test_username')
        # broadinstitute.us
        self.google_us_secure = account.get('google_us_secure')
        self.google_us_username = account.get('google_us_username')
        # test.broadinstitute.us
        self.google_us_test_secure = account.get('google_us_test_secure')
        self.google_us_test_username = account.get('google_us_test_username')

        # gsuite license
        self.gsuite_license = 'Google-Apps-Lite'
        if self.google_license == 'Enterprise':
            self.gsuite_license = 'Google-Apps-Enterprise'

        # domains
        self.domains = {}

        # broadinstitute.org
        if self.google_username:
            self.domains['broadinstitute.org'] = {
                'username': str(self.google_username),
                'secure': self.google_secure,
            }
        if self.google_test_username:
            self.domains['test.broadinstitute.org'] = {
                'username': str(self.google_test_username),
                'secure': self.google_test_secure,
            }
        if self.google_us_username:
            self.domains['broadinstitute.us'] = {
                'username': str(self.google_us_username),
                'secure': self.google_us_secure,
            }
        if self.google_us_test_username:
            self.domains['test.broadinstitute.us'] = {
                'username': str(self.google_us_test_username),
                'secure': self.google_us_test_username,
            }



# -*- coding: utf-8 -*-
"""LDAP Update class for Google."""

import re

from bits.ldap.server import Server
from datetime import date, timedelta


class Google(Server):
    """Google LDAP Update class."""

    def get_default_ous(self):
        """Return list of default ous."""
        # add ous to the list of default_ous
        # these ous do not have any ldap users in them but they need to exist for GCDS
        default_ous = []
        for item in [
            # devices
            'ou=Devices',
            'ou=Admin,ou=Devices',
            'ou=Allow Non-Broad Logins,ou=Devices',
            'ou=Digital Signage,ou=Devices',
            'ou=Loaners,ou=Devices',
            'ou=Meet,ou=Devices',
            'ou=Meet Unmonitored,ou=Devices',
            'ou=Unassigned,ou=Devices',
            # people
            'ou=Insecure,ou=People',
        ]:
            default_ous.append('%s,%s' % (item, self.ldap.base_dn))
        return default_ous

    def get_account_dn(self, account):
        """Return the dn for an other account."""
        # set the base ou
        ou = 'ou=Other Accounts,%s' % (self.ldap.base_dn)

        # check for an ou override
        if account.google_ou:
            parts = account.google_ou.split('/')
            for part in parts:
                ou = 'ou=%s,%s' % (part, ou)

        # otherwise, get the type ou
        else:
            type_ou = self.get_account_type_ou(account)
            if type_ou:
                ou = 'ou=%s,%s' % (type_ou, ou)

        # get primary domain
        domain = None
        if 'primary' in self.ldap.domains:
            domain = self.ldap.domains['primary']

        # get primary domain info
        domain_info = {}
        if domain in account.domains:
            domain_info = account.domains[domain]

        # get secure info
        if domain_info.get('secure'):
            ou = 'ou=Secure,%s' % (ou)

        return 'cn=%s,%s' % (account.displayName, ou)

    def get_account_type_ou(self, account):
        """Return the OU based on the type."""
        if account.type == 'saAccount':
            return 'Admins'
        elif account.type == 'externalCollaborator':
            return 'Collaborators'
        elif account.type == 'roleAccount':
            return 'Role Accounts'
        elif account.type == 'sharedAccount':
            return 'Shared Accounts'
        elif account.type == 'testAccount':
            return 'Test Accounts'

    def get_person_dn(self, person):
        """Return the proper DN for a person."""
        # terminated users who are on hold - suspended
        if person.terminated and person.hold:
            ou = 'ou=Hold,%s' % (self.ldap.base_dn)

        # regular terminated users - suspended
        elif person.terminated:
            ou = 'ou=Suspended,%s' % (self.ldap.base_dn)

        # future hires starting beyond the "start" cutoff - suspended
        elif person.start_date > self.start:
            ou = 'ou=Future,%s' % (self.ldap.base_dn)

        # all other people - active
        else:
            ou = 'ou=People,%s' % (self.ldap.base_dn)

            # get trusted tester ou
            if person.trusted_tester and not person.terminated:
                # ou = 'ou=%s,%s' % (person.trusted_tester, ou)
                ou = 'ou=%s,ou=Trusted Testers,%s' % (person.trusted_tester, ou)

            # otherwise, get workday ou
            else:
                # TEMPORARY for testing
                # if 'test.broadinstitute.' in self.ldap.primary:
                #     ou = self.get_workday_orgunit(person, ou)
                ou = self.get_workday_orgunit(person, ou)

        # create dn
        dn = 'uid=%s,%s' % (person.username, ou)

        return str(dn)

    def get_person(self, person, supervisor_dn):
        """Return a person in the correct format."""
        # get username and email from person object
        username = str(person.username)

        # get email from person objects
        email = str(person.email)
        email_username = str(email.replace('@broadinstitute.org', ''))

        # get nicknames
        nicknames = []
        if person.nicknames:
            for nickname in person.nicknames:
                nicknames.append(str(nickname))

        # get primary and secondary domain from ldap object
        primary = self.ldap.domains.get('primary')
        secondary = self.ldap.domains.get('secondary')

        emails = []

        # set primary domian email
        if primary:
            primary = str(primary)

            # get email address based on broad email (may have alias override)
            email = '%s@%s' % (email_username, primary)
            if email not in emails:
                emails.append(email)

            # get primaryEmail
            primaryEmail = '%s@%s' % (username, primary)
            if primaryEmail not in emails:
                emails.append(primaryEmail)

            # get username/nickname emails for primary domain
            emails = self.get_person_domain_emails(username, nicknames, primary, emails)

        # add secondary domain email
        if secondary:
            secondary = str(secondary)

            # get username/nickname emails for primary domain
            emails = self.get_person_domain_emails(username, nicknames, secondary, emails)

        # convert object to ldap record
        ldap_person = person.to_google(supervisor_dn)

        # overwrite email
        ldap_person['description'] = [str(primaryEmail)]
        ldap_person['email'] = [str(email)]
        ldap_person['mail'] = emails

        return ldap_person

    def get_person_domain_emails(self, username, nicknames, domain, emails=[]):
        """Return the list of domain emails for the user."""
        domain = str(domain)
        username = str(username)

        # get username email
        username_email = '%s@%s' % (username, domain)
        if username_email not in emails:
            emails.append(username_email)

        # get nickname emails
        for nickname in nicknames:
            nickname = str(nickname)
            nickname_email = '%s@%s' % (nickname, domain)
            if nickname_email not in emails:
                emails.append(nickname_email)

        return emails

    def get_workday_orgunit(self, person, ou):
        """Get the orgunit for a workday person."""
        # number of org levels to include
        depth = 2
        for org in person.organization[:depth]:
            org = org.replace(',', '')
            ou = 'ou=%s,%s' % (org, ou)
        return ou

    #
    # Assemled all records into a dict of new entries
    #
    def prepare_entries(self, ldap):
        """Prepare LDAP entries."""
        self.ldap = ldap
        self.prepare_entries_from_people()
        self.prepare_entries_from_accounts()
        self.prepare_entries_from_contacts()
        # self.prepare_entries_from_resources()

    def prepare_entries_from_accounts(self):
        """Prepare entries from accounts."""
        if self.verbose:
            print('  Preparing accounts for Google LDAP...')
        self.accounts = self.prepare_accounts()
        if self.verbose:
            print('  Prepared %s accounts for Google LDAP.' % (len(self.accounts)))
        # add contacts
        for dn in self.accounts:
            self.new_entries[dn] = self.accounts[dn]

    def prepare_entries_from_contacts(self):
        """Prepare entries from contacts."""
        if self.verbose:
            print('  Preparing contacts for Google LDAP...')
        self.contacts = self.prepare_contacts()
        if self.verbose:
            print('  Prepared %s contacts for Google LDAP.' % (len(self.contacts)))
        # add contacts
        for dn in self.contacts:
            self.new_entries[dn] = self.contacts[dn]

    def prepare_entries_from_people(self):
        """Prepare entries from people."""
        if self.verbose:
            print('  Preparing people for Google LDAP...')
        self.new_entries = self.prepare_people()
        if self.verbose:
            print('  Prepared %s people for Google LDAP.' % (len(self.new_entries)))

    def prepare_entries_from_resources(self):
        """Prepare entries from resources."""
        if self.verbose:
            print('  Preparing resources for Google LDAP...')
        self.resources = self.prepare_resources()
        if self.verbose:
            print('  Prepared %s resources for Google LDAP.' % (len(self.resources)))
        # add resources
        for dn in self.resources:
            self.new_entries[dn] = self.resources[dn]

    #
    # Prepare actual resources for LDAP
    #
    def prepare_accounts(self):
        """Prepare other accounts from bitsdb."""
        accounts = {}
        for account in self.update.accounts_records:
            # get dn
            dn = self.get_account_dn(account)

            # get emails
            emails = []

            # get primary email
            primary = self.ldap.domains.get('primary')
            primary_info = account.domains.get(primary, {})
            email = primary_info.get('username')
            if email:
                emails.append(str(email))
            else:
                # skip users that have no email
                continue

            # check for sub domains
            subdomain = False
            if '.%s' % (primary) in email:
                subdomain = True

            # check for users with primary email in secondary domain
            secondary_user = False
            if primary not in email:
                secondary_user = True

            # get secondary email
            secondary = self.ldap.domains.get('secondary')
            if email and secondary and not secondary_user and not subdomain:
                # create secondary email by replacing primary domain with secondary
                secondary_email = email.replace(primary, secondary)
                if secondary_email:
                    emails.append(str(secondary_email))

            # create record
            record = {
                'cn': [account.displayName],
                'carLicense': [account.gsuite_license],
                'description': [email],
                'givenName': [account.first_name],
                'co': ['Broad Institute of MIT and Harvard'],
                'mail': emails,
                'o': ['Broad Institute'],
                'objectClass': [
                    'top',
                    'inetOrgPerson',
                    'extensibleObject',
                ],
                'sn': [account.last_name],
            }

            # only add if the account has emails for this domain
            if email and emails:
                accounts[dn] = record

        return accounts

    def prepare_contacts(self):
        """Prepare contact records for Google LDAP."""
        contacts = {}
        for oid in self.update.shared_contacts:
            contact = self.update.shared_contacts[oid]

            # set cn and dn
            cn = str('%s %s' % (contact['first_name'], contact['last_name']))
            ou = str('ou=Contacts,ou=Shared Contacts,%s' % (self.ldap.base_dn))
            dn = str('cn=%s,%s' % (cn, ou))

            # get building info
            building = None
            if 'building' in contact and contact['building']:
                building = self.update.buildings[contact['building']]

            # create record
            record = {
                'cn': [cn],
                'description': [str(contact['email'])],
                'email': [str(contact['email'])],
                'givenName': [str(contact['first_name'])],
                'mail': [str(contact['email'])],
                'o': [str(contact['org'])],
                'objectClass': [
                    'top',
                    'inetOrgPerson',
                    'extensibleObject',
                ],
                'sn': [str(contact['last_name'])],
            }

            # building
            if building:
                record['l'] = [building['city']]
                record['postalCode'] = [building['zip']]
                record['st'] = [building['state']]
                record['street'] = [building['street_address']]

            # telephone
            if contact.get('phone'):
                record['telephoneNumber'] = [str(contact['phone'])]

            # add contact to dictionar
            contacts[dn] = record

        return contacts

    def prepare_people(self):
        """Prepare people records for Google LDAP."""
        # get dates
        date_today = date.today()
        start_date = date_today + timedelta(days=14)
        end_date = date_today - timedelta(days=90)

        # get date strings
        # today = date_today.strftime('%Y-%m-%d')
        self.start = start_date.strftime('%Y-%m-%d')
        self.end = end_date.strftime('%Y-%m-%d')

        people = {}
        for pid in self.update.people_records:
            # get person object
            person = self.update.people_records[pid]

            # skip people who are future hires
            # if person.start_date > self.start:
            #     continue

            # skip people are are past 90 days termination
            if not person.hold and (person.end_date and person.end_date <= self.end):
                continue

            # for broadinstitute.us domains, skip employees
            primary_domain = self.ldap.domains.get('primary')
            if primary_domain and re.search('broadinstitute.us$', primary_domain):
                if person.type in ['Core Member', 'Employee']:
                    continue

            dn = self.get_person_dn(person)
            # get supervisor object
            supervisor_dn = None
            if person.supervisor_id:
                supervisor = self.update.people_records[person.supervisor_id]
                supervisor_dn = self.get_person_dn(supervisor)
            # get person in ldap format
            people[dn] = self.get_person(person, supervisor_dn)
        return people

    def prepare_resources(self):
        """Prepare calendar resources for Google LDAP."""
        resources = {}
        ou = 'ou=Resources,ou=Shared Contacts,%s' % (self.ldap.base_dn)

        # get buildings by code
        buildings = {}
        for bid in self.update.buildings:
            building = self.update.buildings[bid]
            code = building['code'].lower()
            buildings[code] = building

        # create resource records
        for resource in self.update.resources:
            # only include conference rooms

            if 'resourceType' not in resource:
                print('ERROR: No Resource Type: %s' % (resource))
                continue

            if not re.search('Conference Room', resource['resourceType']):
                continue

            # get dn
            dn = 'cn=%s,%s' % (resource['resourceName'], ou)

            # get room and name
            building_room = str(resource['resourceDescription'].split(',')[0])
            room = str(building_room.split('-')[1].strip())
            name = str(resource['resourceName'].split(' (')[0].strip())

            # create record
            record = {
                'cn': [str(resource['resourceName'])],
                'description': [str(resource['resourceDescription'])],
                'email': [str(resource['resourceEmail'])],
                'employeeNumber': [str(resource['resourceId'])],
                'givenName': [name],
                # 'l': ['Cambridge'],
                'objectClass': [
                    'top',
                    'inetOrgPerson',
                    'extensibleObject',
                ],
                # 'postalCode': ['02142'],
                'roomNumber': [room],
                'sn': [str(resource['resourceType'])],
                # 'st': ['MA'],
                # 'street': ['75 Ames Street'],
            }

            # get building
            building_code = resource.get('buildingId')
            if building_code in buildings:
                building = buildings[building_code]
                record['l'] = [building['city']]
                record['postalCode'] = [building['zip']]
                record['st'] = [building['state']]
                record['street'] = [building['street_address']]

            # save record
            resources[dn] = record

        return resources



# -*- coding: utf-8 -*-
"""LDAP Update class for Active Directory."""

import datetime
import pytz
import re
import time

from bits.ldap import LDAP as LDAPBase


class AD(LDAPBase):
    """AD LDAP Update class."""

    def __init__(
            self,
            uri,
            bind_dn,
            bind_pw,
            base_dn,
            server_type=None,
            domains=None,
            bitsdb_key=None,
            verbose=False):
        """Initialize a class instance."""
        LDAPBase.__init__(self, uri, bind_dn, bind_pw, base_dn, server_type, domains, bitsdb_key, verbose)

        self.group_attributes = [
            'cn',
            'dSCorePropagationData',
            'description',
            'distinguishedName',
            'gidNumber',
            'groupType',
            'info',
            'instanceType',
            'managedBy',
            'member',
            # 'member;range=0-1499',
            'name',
            'objectCategory',
            'objectClass',
            # 'objectGUID',
            # 'objectSid',
            'sAMAccountName',
            'sAMAccountType',
            'uSNChanged',
            'uSNCreated',
            'whenChanged',
            'whenCreated',
        ]

        self.user_attributes = [
            'accountExpires',
            # 'adminCount',
            # 'badPasswordTime',
            # 'badPwdCount',
            'buildingName',
            'cn',
            'co',
            # 'codePage',
            'company',
            # 'countryCode',
            # 'dSCorePropagationData',
            'department',
            'departmentNumber',
            'description',
            'directReports',
            'displayName',
            'distinguishedName',
            'division',
            'employeeNumber',
            'employeeType',
            'extensionAttribute11',
            'gidNumber',
            'givenName',
            'homeDirectory',
            'homeDrive',
            'info',
            'initials',
            # 'instanceType',
            # 'lastLogoff',
            # 'lastLogon',
            # 'lastLogonTimestamp',
            'l',
            # 'lockoutTime',
            'loginShell',
            # 'logonCount',
            # 'logonHours',
            'mail',
            'managedObjects',
            'manager',
            'maxPwdAge',
            'member',
            'memberOf',
            'msRADIUSCallbackNumber',
            'msRADIUSFramedIPAddress',
            'msRADIUSServiceType',
            # 'msTSExpireDate',
            # 'msTSLicenseVersion',
            # 'msTSLicenseVersion2',
            # 'msTSLicenseVersion3',
            # 'msTSManagingLS',
            'name',
            'o',
            'objectCategory',
            # 'objectClass',
            # 'objectGUID',
            # 'objectSid',
            'physicalDeliveryOfficeName',
            'postalCode',
            'primaryGroupID',
            'profilePath',
            'pwdLastSet',
            'sAMAccountName',
            'sAMAccountType',
            # 'scriptPath'
            'sn',
            'st',
            'streetAddress',
            'telephoneNumber',
            # 'terminalServer'
            'title',
            # 'uSNChanged',
            # 'uSNCreated',
            'uid',
            'uidNumber',
            'unixHomeDirectory',
            'userAccountControl',
            # 'userCertificate',
            'userPrincipalName',
            'whenChanged',
            'whenCreated',
        ]

        self.groups = {}
        self.google_users = {}
        self.users = {}
        self.users_by_dn = {}

    #
    # Helpers
    #
    def get_expire_date(self, last, maxDays=None):
        """Convert pwd last set date to a datetime."""
        if not last:
            return datetime.datetime(year=1970, month=1, day=1)
        if not maxDays:
            maxDays = 90
        epoch = datetime.datetime(year=1601, month=1, day=1)
        last_set = datetime.timedelta(seconds=last/10**7)
        max_age = datetime.timedelta(days=maxDays)
        return epoch + last_set + max_age

    def get_pwd_last_set_date(self, last):
        """Convert pwd last set date to an expiration date."""
        epoch = datetime.datetime(year=1601, month=1, day=1)
        pwdlastset = datetime.timedelta(seconds=last/10**7)
        return epoch + pwdlastset

    #
    # Groups
    #
    def check_group(self, groupname):
        """Check if an Active Directory group exists."""
        try:
            if self.get_group(groupname):
                return True
        except Exception:
            return False

    def get_group(self, groupname, bytes_mode=True):
        """Return a single group."""
        filterstr = '(sAMAccountName={})'.format(groupname)

        # bytes mode
        if bytes_mode:
            entries = self.get_entries(filterstr, self.group_attributes)
            for dn in entries:
                return entries[dn]

        # unicode strings mode
        entries = self.get_entries(filterstr, self.group_attributes)
        for dn in entries:
            return self.convert_entry_to_strings(entries[dn])

    def get_groups(self, attrlist=None, bytes_mode=True, key='sAMAccountName'):
        """Return a dict of groups by groupname."""
        entries = self.get_groups_by_dn(attrlist=attrlist, bytes_mode=bytes_mode)
        groups = {}
        for dn in entries:
            e = entries[dn]
            k = e[key][0]
            if isinstance(k, bytes):
                k = k.decode('utf-8')
            groups[k] = e
        return groups

    def get_groups_by_dn(self, attrlist=None, bytes_mode=True):
        """Return a dict of groups from Active Directory."""
        filterstr = '(&(objectCategory=group)(sAMAccountName=*))'

        if not attrlist:
            attrlist = self.user_attributes

        # bytes mode
        if bytes_mode:
            return self.get_entries(filterstr, self.user_attributes)

        # unicode strings mode
        entries = self.get_entries(filterstr, self.user_attributes)
        groups = {}
        for dn in entries:
            groups[dn] = self.convert_entry_to_strings(entries[dn])
        return groups

    def get_groups_by_gid(self, bytes_mode=True):
        """Return a dict of groups by gid."""
        groups = self.get_groups(bytes_mode=bytes_mode)
        gid_groups = {}
        for groupname in groups:
            g = groups[groupname]
            if 'gidNumber' in g and g['gidNumber'] and int(g['gidNumber'][0]) >= 0:
                gid = int(g['gidNumber'][0])
                if gid in gid_groups:
                    gid_groups[gid] = [groupname]
                else:
                    gid_groups[gid] = [groupname]
        return gid_groups

    def get_groups_by_username(self, bytes_mode=True):
        """Return a dict of groups by username."""
        groups = self.get_groups_for_nis()
        user_groups = {}
        for groupname in groups:
            g = groups[groupname]
            members = g['users']
            for m in members:
                if not m:
                    continue
                if m not in user_groups:
                    user_groups[m] = []
                user_groups[m].append(groupname)

        return user_groups

    def get_groups_for_nis(self, bytes_mode=True, people=None):
        """Return a dict of groups for NIS."""
        groups = self.get_groups(bytes_mode=False)
        users = self.get_users_by_dn(bytes_mode=False)

        nis_groups = {}
        notfound = []
        for groupname in groups:
            g = groups[groupname]

            # only get groups that have a gid
            if 'gidNumber' not in g or int(g['gidNumber'][0]) < 0:
                continue

            gid = int(g['gidNumber'][0])

            members = []

            # see if the group has members:
            if 'member' in g:

                # cycle through the members
                for m in g['member']:

                    # skip users without a dn
                    if not m:
                        continue

                    # check if user exists in AD
                    if m in users:
                        u = users[m]

                        # set the username
                        username = u['sAMAccountName'][0].lower()

                        if people and username in people and people[username]['terminated']:
                            continue

                        # add the username to the list of members
                        members.append(username)

                    else:
                        print('ERROR: AD user not found!: {} ({})'.format(m, groupname))
                        notfound.append(m)

            data = {
                'name': groupname,
                'gid': gid,
                'password': '*',
                'users': members,
            }

            nis_groups[groupname] = data

        return nis_groups

    def update_groups(self, auth):
        """Update groups in Active Directory."""
        b = auth.bitsdbapi()
        ad_group_syncs = b.get_ad_group_syncs()

        g = auth.google()
        g.auth_service_account(g.scopes, g.subject)

        if auth.verbose:
            print("Getting Google Users...")
        self.google_users = g.directory().get_users_dict()

        default_users = {
            # "engineering": [
            #     "devopsci",
            #     "itweb",
            #     "rpmbuild",
            #     "svcnow",
            # ],
        }

        if auth.verbose:
            print("Getting AD Users...")
        self.get_users(bytes_mode=False)

        print('Updating AD group members from Google Groups...')
        for sync in ad_group_syncs:
            ad_group_name = sync["ad_group_name"]
            google_group_email = sync["google_group_email"]
            if auth.verbose:
                print(f"\nUpdating AD group \"{ad_group_name}\" from {google_group_email}...")

            members = default_users.get(ad_group_name, [])
            members.extend(self.get_google_group_members(google_group_email, auth))
            self.update_group_members(ad_group_name, members)

    #
    # Members
    #
    def add_group_member(self, groupname, username):
        """Add a user to an Active Directory group."""
        group = self.get_group(groupname, bytes_mode=False)
        if not group:
            print('ERROR: No such group: {}'.format(groupname))
            return

        user = self.get_user(username, bytes_mode=False)
        if not user:
            print('ERROR: No such user: {}'.format(username))
            return

        # get dns
        groupDn = group['distinguishedName'][0]
        userDn = user['distinguishedName'][0]

        # check if group has any members
        if 'member' not in group:
            group['member'] = []

        # check if user is already in group
        if userDn in group['member']:
            return True

        modlist = [(self.ldapmodule.MOD_ADD, 'member', userDn.encode('utf-8'))]
        return self.ldapobject.modify_s(groupDn, modlist)

    def get_google_group_members(self, groupname, auth):
        """Return the members from a Google Group for syncing with AD."""
        g = auth.google()
        g.auth_service_account(g.scopes, g.subject)

        # add domain name
        group_email = groupname
        if "@" not in groupname:
            group_email = f"{groupname}@broadinstitute.org"

        # get members
        members = g.directory().get_derived_members(group_email)

        # create users list
        users = []
        for m in members:
            email = m['email']
            uid = m["id"]

            if uid in self.google_users:
                email = self.google_users[uid]["primaryEmail"]

            username = email.replace("@broadinstitute.org", "")
            users.append(username)

        return sorted(set(users))

    def remove_group_member(self, groupname, username):
        """Remove a user from an Active Directory group."""
        group = self.get_group(groupname, bytes_mode=False)
        if not group:
            print('ERROR: No such group: {}'.format(groupname))
            return

        user = self.get_user(username, bytes_mode=False)
        if not user:
            print('ERROR: No such user: {}'.format(username))
            return

        # get dns
        groupDn = group['distinguishedName'][0]
        userDn = user['distinguishedName'][0]

        # check if group has any members
        if 'member' not in group:
            group['member'] = []

        # check if user is already in group
        if userDn not in group['member']:
            return True

        modlist = [(self.ldapmodule.MOD_DELETE, 'member', userDn.encode('utf-8'))]
        return self.ldapobject.modify_s(groupDn, modlist)

    def update_group_members(self, groupname, groupusers):
        """Update the members of an Active Directory groups."""
        group = self.get_group(groupname, bytes_mode=False)

        if 'member' not in group:
            members = []
        else:
            members = group['member']

        # find members to add
        add = []
        for username in sorted(groupusers):
            if username not in self.users:
                if self.verbose:
                    print('WARNING: AD username not found [%s]: %s' % (
                        groupname,
                        username,
                    ))
                continue

            user = self.users[username]
            dn = user['distinguishedName'][0]
            if dn not in members:
                add.append(username)

        # find members to remove
        remove = []
        for dn in sorted(members):
            if dn not in self.users_by_dn:
                print('ERROR: AD user DN not found [%s]: %s' % (
                    groupname,
                    dn
                ))
                continue

            user = self.users_by_dn[dn]
            username = user['sAMAccountName'][0]
            if username not in groupusers:
                remove.append(username)

        if add or remove:
            print('Updating AD group: %s' % (groupname))

        # add users
        if add:
            print('  Adding users:')
            for username in sorted(add):
                print('   + %s' % (username))
                self.add_group_member(groupname, username)

        # remove users
        if remove:
            print('  Removing users:')
            for username in sorted(remove):
                print('   - %s' % (username))
                self.remove_group_member(groupname, username)

    #
    # OUs
    #
    def get_ous(self, bytes_mode=True):
        """Return a dict of ous from Active Directory."""
        attrlist = ['ou']
        filterstr = '(ou=*)'

        # bytes mode
        if bytes_mode:
            return self.get_entries(filterstr, attrlist)

        # unicode strings mode
        entries = self.get_entries(filterstr, attrlist)
        ous = {}
        for dn in entries:
            ous[dn] = self.convert_entry_to_strings(entries[dn])
        return ous

    #
    # Users
    #
    def check_user(self, username):
        """Check if an Active Directory user exists."""
        try:
            if self.get_user(username):
                return True
        except Exception:
            return False

    def create_user(self, username):
        """Create a user in Active Directory."""
        """TODO"""

    def delete_user(self, username):
        """Delete a user in Active Directory."""
        user = self.get_user(username, bytes_mode=False)
        if not user:
            print('ERROR: No such user: {}'.format(username))
            return

        # get dn
        dn = user['distinguishedName'][0]

        # delete user
        self.delete_entry(dn)

        return user

    def get_user(self, username, bytes_mode=True):
        """Return a single user."""
        filterstr = '(sAMAccountName={})'.format(username)

        # bytes mode
        if bytes_mode:
            entries = self.get_entries(filterstr, self.user_attributes)
            for dn in entries:
                return entries[dn]

        # unicode strings mode
        entries = self.get_entries(filterstr, self.user_attributes)
        for dn in entries:
            return self.convert_entry_to_strings(entries[dn])

    def get_users(self, attrlist=None, bytes_mode=True, key='sAMAccountName'):
        """Return a dict of users by username."""
        entries = self.get_users_by_dn(attrlist=attrlist, bytes_mode=bytes_mode)
        users = {}
        for dn in entries:
            e = entries[dn]
            k = e[key][0]
            if isinstance(k, bytes):
                k = k.decode('utf-8')
            users[k] = e
        self.users = users
        return users

    def get_users_by_dn(self, attrlist=None, bytes_mode=True):
        """Return a dict of users from Active Directory."""
        filterstr = '(&(objectCategory=person)(sAMAccountName=*))'

        if not attrlist:
            attrlist = self.user_attributes

        # bytes mode
        if bytes_mode:
            return self.get_entries(filterstr, attrlist)

        # unicode strings mode
        entries = self.get_entries(filterstr, attrlist)
        users = {}
        for dn in entries:
            users[dn] = self.convert_entry_to_strings(entries[dn])
        self.users_by_dn = users
        return users

    def get_users_for_nis(self):
        """Return a dict of users for NIS."""
        users = self.get_users(bytes_mode=False)
        nis_users = {}
        for username in sorted(users):
            if not username:
                continue

            u = users[username]

            if not u:
                print('ERROR: AD user not found: %s' % (username))
                continue

            if 'uidNumber' not in u or int(u['uidNumber'][0]) < 0:
                continue

            if 'gidNumber' not in u or int(u['gidNumber'][0]) < 0:
                continue

            if 'loginShell' not in u or not str(u['loginShell'][0]):
                continue

            if 'unixHomeDirectory' not in u or not str(u['unixHomeDirectory'][0]):
                continue

            uid = int(u['uidNumber'][0])
            gid = int(u['gidNumber'][0])
            home_dir = u['unixHomeDirectory'][0]
            shell = u['loginShell'][0]
            gecos = u['displayName'][0]

            user = {
                'username': username,
                'uid': uid,
                'gid': gid,
                'gecos': gecos,
                'password': '*',
                'shell': shell,
                'home_dir': home_dir,
            }

            nis_users[username] = user

        return nis_users

    def lock_user(self, username):
        """Lock a user in Active Directory."""
        user = self.get_user(username, bytes_mode=False)
        if not user:
            print('ERROR: No such user: {}'.format(username))
            return

        dn = user['distinguishedName'][0]

        # lock account
        locked = False
        if user['userAccountControl'][0] != '514':
            modlist = [(self.ldapmodule.MOD_REPLACE, 'userAccountControl', b'514')]
            try:
                self.ldapobject.modify_s(dn, modlist)
                locked = True
            except Exception as e:
                print(type(e))
                print(e)
                print('ERROR: Failed to unlock AD user: {}'.format(username))

        # move user
        moved = False
        if not re.search('Disabled Accounts', dn):
            self.move_user(username, 'OU=Disabled Accounts,DC=broad,DC=mit,DC=edu')
            moved = True

        if locked or moved:
            return True

    def move_user(self, username, newParent):
        """Move a user in Active Directory."""
        user = self.get_user(username, bytes_mode=False)
        if not user:
            print('ERROR: No such user: {}'.format(username))
            return

        dn = user['distinguishedName'][0]
        rdn = 'CN={}'.format(user['cn'][0])
        return self.ldapobject.rename_s(dn, rdn, newParent)

    def prepare_user(self):
        """Prepare a single user for in Active Directory."""
        """TODO"""

    def rename_user(self, username, newrdn):
        """Rename a user in Active Directory."""
        user = self.get_user(username, bytes_mode=False)
        if not user:
            print('ERROR: No such user: {}'.format(username))
            return

        dn = user['distinguishedName'][0]
        return self.ldapobject.rename_s(dn, newrdn)

    def set_password(self, username, password):
        """Set a user's password in Active Directory."""
        user = self.get_user(username, bytes_mode=False)
        if not user:
            print('ERROR: No such user: {}'.format(username))
            return

        dn = user['distinguishedName'][0]
        adPassword = '"{}"'.format(password).encode('utf-16-le')
        modlist = [(self.ldapmodule.MOD_REPLACE, 'unicodePwd', [adPassword])]
        return self.ldapobject.modify_s(dn, modlist)

    def unlock_user(self, username):
        """Unlock a user in Active Directory."""
        user = self.get_user(username, bytes_mode=False)
        if not user:
            print('ERROR: No such user: {}'.format(username))
            return

        dn = user['distinguishedName'][0]

        # check if account needs to be unlocked
        unlocked = False
        if user['userAccountControl'][0] != '512':
            modlist = [
                (self.ldapmodule.MOD_REPLACE, 'userAccountControl', b'512'),
                (self.ldapmodule.MOD_REPLACE, 'loginShell', b'/bin/bash')
            ]
            try:
                self.ldapobject.modify_s(dn, modlist)
                unlocked = True
            except Exception as e:
                print(type(e))
                print(e)
                print('ERROR: Failed to unlock AD user: {}'.format(username))

        # check if account needs to be moved
        moved = False
        if not re.search('UserOU', dn):
            try:
                self.move_user(username, 'OU=UserOU,DC=broad,DC=mit,DC=edu')
                moved = True
            except Exception as e:
                print(type(e))
                print(e)
                print('ERROR: Failed to move AD user: {}'.format(username))

        if unlocked or moved:
            return True

    # def update_user(self):
    #     """Update a single user in Active Directory."""
    #     """TODO"""

    # def update_users(self):
    #     """Update users in Active Directory."""
    #     """TODO"""

    #
    # Update Users
    #
    def compare_user(self, old, new):
        """Compare a single user for update."""
        exclude = [
            'distinguishedName',
            'directReports',
            'homeDrive',
            'info',
            'managedObjects',
            'memberOf',
            'msRADIUSCallbackNumber',
            'msRADIUSFramedIPAddress',
            'msRADIUSServiceType',
            # 'name',
            'pwdLastSet',
            'whenChanged',
            'whenCreated',
        ]

        # copy dicts
        new = dict(new)
        old = dict(old)

        # reverse organization for comparisions (FIFO)
        if old.get("o"):
            old["o"].reverse()

        output = []
        for key in set(list(old) + list(new)):
            if key in exclude:
                if key in old:
                    del old[key]
                if key in new:
                    del new[key]
                continue

            n = new.get(key, [None])
            o = old.get(key, [None])

            if n != o:
                output.append('{}: {} -> {}'.format(key, o, n))

        ldif = None
        if output:
            ldif = self.ldapmodule.modlist.modifyModlist(old, new)

        return ldif, output

    def get_person_dn(self, person):
        """Return the DN of a person."""
        # print(person)
        dn = None
        return dn

    def prepare_person(self, p, u):
        """Prepare a single person for Active Directory."""
        # print(json.dumps(p, indent=2, sort_keys=True))
        username = p["username"]
        username_email = f"{username}@broadinstitute.org"
        terminated = p["terminated"]

        dn = self.get_person_dn(p)

        # set the date when accountExpires
        accountExpires = "0"
        if p.get("end_date"):
            date_format = "%Y-%m-%d %H:%M:%S"
            end_date = p["end_date"]
            est = pytz.timezone("US/Eastern")
            # set expire time to 9am GMT 4am EST on day after termination date
            expire_string = f"{end_date} 05:00:00"
            expire_time = datetime.datetime.strptime(expire_string, date_format) + datetime.timedelta(days=1)
            # handle daylight savings time
            expire_time -= est.localize(expire_time).dst()
            # convert to unix timestamp
            unix_expire = int(time.mktime(expire_time.timetuple()))
            # conver to AD timestamp
            accountExpires = str((unix_expire + 11644473600) * 10000000)

        # set the address fields
        street_address = None
        city = None
        state = None
        postal_code = None
        country = None
        if p.get("address"):
            street_address = p["address"]["street"]
            city = p["address"]["city"]
            state = p["address"]["state"]
            postal_code = p["address"]["zip_code"]
            country = p["address"]["country"]

        # set the company
        company = p["home_institution_ascii"]
        if len(company) > 64:
            company = company[0:64]

        # set the department:
        department = None
        if p.get("org_unit"):
            org_unit = p["org_unit"]
            org_unit = org_unit.replace("Administration > ", "Admin > ")
            org_unit = org_unit.replace("Research > ", "")
            org_unit = org_unit.replace("Specialized Service Facilities > ", "")

            department = org_unit.split(" > ")[-1]
            department = department.replace("&", "-").replace("/", "-")

        # set the department numb
        department_number = p["department_id"]

        # set the description
        description = p["worker_sub_type"]

        # set the division
        division = None
        if p.get("org_unit"):
            division = p["org_unit"]

        # set the initials
        initials = None

        # set the name
        first_name = p["first_name_ascii"]
        last_name = p["last_name_ascii"]
        full_name = p["full_name_ascii"]

        # set the email
        email = p["email"]

        # set the employee number
        employee_number = p["emplid"]

        # set the employee type
        employee_type = p["worker_type"]

        # set the homeDirectory and homeDrive
        homeDirectory = f"\\\\krypton\\{username}"
        homeDrive = "I:"

        # set the manager
        manager = None
        if "manager_dn" in p and p["manager_dn"]:
            manager = p["manager_dn"]

        # set the orgaization
        organization = [None]
        if p.get("department"):
            organization = p["department"][1:]

        # set objectCategory
        objectCategory = "CN=Person,CN=Schema,CN=Configuration,DC=broad,DC=mit,DC=edu"

        # set the physicalDeliveryOfficeName
        # building_name = None
        physicalDeliveryOfficeName = None
        if p.get("desk"):
            # building_name = p["desk"].split("-")[0]
            physicalDeliveryOfficeName = p["desk"]

        # set primaryGroupID
        primaryGroupID = "513"

        # set samAccountType
        sAMAccountType = "805306368"

        # set the telephoneNumber
        telephoneNumber = None
        if p.get("primary_work_phone"):
            telephoneNumber = p["primary_work_phone"]

        # set the title
        title = p["title"]

        # set userAccountControl and extensionAttribute11
        extensionAttribute11 = "SciQuest"
        userAccountControl = "512"
        if terminated:
            extensionAttribute11 = None
            userAccountControl = "514"

        # get nis account info
        loginShell = None
        if u and u.get("loginShell"):
            loginShell = u["loginShell"][0]
            if terminated:
                loginShell = "/bin/false"
            elif loginShell == "/bin/false":
                loginShell = "/bin/bash"
        elif not terminated:
            print(f"WARNING: No login shell for user: {username}")

        gidNumber = None
        if u and u.get("gidNumber"):
            gidNumber = u["gidNumber"][0]
        elif not terminated:
            print(f"WARNING: No gid number for user: {username}")

        uidNumber = None
        if u and u.get("uidNumber"):
            uidNumber = u["uidNumber"][0]
        elif not terminated:
            print(f"WARNING: No uid number for user: {username}")

        unixHomeDirectory = None
        if u and u.get("unixHomeDirectory"):
            unixHomeDirectory = f"/home/unix/{username}"
        elif not terminated:
            print(f"WARNING: No UNIX home directory for user: {username}")

        entry = {
            "accountExpires": [accountExpires],
            # "buildingName": [building_name],
            "cn": [username],
            "co": [country],
            "company": [company],
            "department": [department],
            "departmentNumber": [department_number],
            "description": [description],
            "displayName": [full_name],
            "distinguishedName": [dn],
            "division": [division],
            "employeeNumber": [employee_number],
            "employeeType": [employee_type],
            "extensionAttribute11": [extensionAttribute11],
            "gidNumber": [gidNumber],
            "givenName": [first_name],
            "homeDirectory": [homeDirectory],
            "homeDrive": [homeDrive],
            "initials": [initials],
            "l": [city],
            "loginShell": [loginShell],
            "mail": [email],
            "manager": [manager],
            "name": [username],
            "o": organization,
            "objectCategory": [objectCategory],
            "physicalDeliveryOfficeName": [physicalDeliveryOfficeName],
            "postalCode": [postal_code],
            "primaryGroupID": [primaryGroupID],
            "sAMAccountName": [username],
            "sAMAccountType": [sAMAccountType],
            "sn": [last_name],
            "st": [state],
            "streetAddress": [street_address],
            "telephoneNumber": [telephoneNumber],
            "title": [title],
            "uid": [username],
            "uidNumber": [uidNumber],
            "unixHomeDirectory": [unixHomeDirectory],
            "userAccountControl": [userAccountControl],
            "userPrincipalName": [username_email],
            # "userPrincipalName": [email],
        }
        if username == b'oleary':
            entry["userPrincipalName"] = [email]
        return entry

    def prepare_people(self, people, users):
        """Prepare people data for AD."""
        data = {}
        for emplid in people:
            p = people[emplid].to_dict()

            # get manager
            if p['manager_id'] and p['manager_id'] in people:
                manager_id = p['manager_id']
                manager_username = people[manager_id].to_dict()['username']
                if manager_username in users:
                    p['manager_dn'] = users[manager_username]['distinguishedName'][0]

            username = p['username']
            u = users.get(username)
            entry = self.prepare_person(p, u)
            data[username] = entry
        return data

    def prepare_users(self, auth, users):
        """Prepare user data for AD."""
        g = auth.google()
        g.auth_service_account(g.scopes)

        project = 'broad-bitsdb-firestore'
        collection = 'people_people'

        firestore = g.firestore(project)
        print('Getting people from Firestore...')
        people = firestore.get_docs_dict(collection)
        print('Found {} people.'.format(len(people)))

        # m = auth.mongo()
        # other_accounts = m.getOtherAccounts(key='ad_username')

        people_data = self.prepare_people(people, users)

        return people_data

    def update_users(self, auth):
        """Update users in AD."""
        print("Getting Users from AD...")
        users = self.get_users()
        print("Found {} AD users.".format(len(users)))

        data = self.prepare_users(auth, users)

        add = []
        delete = []
        update = {}

        # find users to update
        for username in sorted(users):
            if username in data:
                new = self.convert_entry(data[username])
                old = users[username]
                ldif, output = self.compare_user(old, new)
                if output:
                    dn = old["distinguishedName"][0].decode("utf-8")
                    print("\nUpdates for: {} [{}]".format(username, dn))
                    print("   * " + "\n   * ".join(output))
                    dn = old["distinguishedName"][0].decode("utf-8")
                    update[dn] = ldif

        print("Add: %s, Delete: %s, Update: %s" % (
            len(add),
            len(delete),
            len(update),
        ))

        # perform updates
        if update:
            print("\nPerforming updates...")
        for dn in sorted(update):
            ldif = update[dn]
            try:
                self.ldapobject.modify_s(dn, ldif)
            except Exception as e:
                print("\nERROR Updating: {}".format(dn))
                print(e)
                print(ldif)
        if update:
            print("Done.")



# -*- coding: utf-8 -*-
"""LDAP class file."""
from __future__ import absolute_import

import json
import ldap as ldapmodule
from ldap import modlist

# from .ad import AD


class LDAP(object):
    """LDAP class."""

    # default page size (for Active Directory servers)
    pagesize = 1000

    # Ignore errors related to self-signed certificates
    ldapmodule.set_option(ldapmodule.OPT_X_TLS_REQUIRE_CERT, ldapmodule.OPT_X_TLS_ALLOW)

    # Don't follow referrals
    ldapmodule.set_option(ldapmodule.OPT_REFERRALS, 0)

    def __init__(
        self,
        uri,
        bind_dn,
        bind_pw,
        base_dn,
        server_type=None,
        domains=None,
        bitsdb_key=None,
        verbose=False
    ):
        """Initialize an instance."""
        self.base_dn = base_dn
        self.bind_dn = bind_dn
        self.bind_pw = bind_pw
        self.bitsdb_key = bitsdb_key
        self.domains = domains
        self.server_type = server_type
        self.uri = uri
        self.verbose = verbose

        # ldap module
        self.ldapmodule = ldapmodule
        self.modlist = modlist

        # Initialize the LDAP connection
        self.ldapobject = ldapmodule.initialize(uri)
        if self.verbose:
            print('  Connected to LDAP server: %s' % (uri))

        # set the protocol version to version 3
        self.ldapobject.protocol_version = ldapmodule.VERSION3

        # Bind to LDAP
        self.ldapobject.simple_bind_s(bind_dn, bind_pw)
        if self.verbose:
            print('  Bound to LDAP server as: %s' % (bind_dn))

        # get domains
        self.primary = None
        self.secondary = None
        if self.domains:
            if 'primary' in self.domains:
                self.primary = self.domains['primary']
            if 'secondary' in self.domains:
                self.secondary = self.domains['secondary']

    # def ad(self):
    #     """Return an AD instance."""
    #     return AD(self)

    def convert_entry(self, entry):
        """Convert a single entry to bytes instead of string."""
        for key in entry:
            items = []
            for i in entry[key]:
                if isinstance(i, str):
                    i = i.encode('utf-8')
                items.append(i)
            entry[key] = items
        return entry

    def convert_entry_to_strings(self, entry):
        """Convert a single entry to strings instead of bytes."""
        for key in entry:
            items = []
            for i in entry[key]:
                if isinstance(i, bytes):
                    i = i.decode('utf-8')
                items.append(i)
            entry[key] = items
        return entry

    def get_entries(self, filterstr='(objectClass=*)', attrlist=None, pagesize=None):
        """Return LDAP entries as a dict by dn."""
        base = self.base_dn
        scope = ldapmodule.SCOPE_SUBTREE
        results = self.search(base, scope, filterstr, attrlist, pagesize)
        entries = {}
        for dn, entry in results:
            if not dn:
                continue
            entries[dn] = entry
        return entries

    def get_ous(self):
        """Return a list of OUs."""
        return self.get_entries('(objectClass=organizationalUnit)')

    def search(self, base, scope, filterstr='(objectClass=*)', attrlist=None, pagesize=None):
        """Perform an LDAP search."""
        # for ad servers, use simple paged resuls
        if self.server_type == 'ad':
            return self.get_paged_results(base, scope, filterstr, attrlist, pagesize)

        return self.ldapobject.search_ext_s(
            base=base,
            scope=scope,
            filterstr=filterstr,
            attrlist=attrlist,
            # attrsonly=0,
            # serverctrls=None,
            # clientctrls=None,
            # timeout=-1,
            # sizelimit=0,
        )

    def get_ldap_controls(self, pagesize=None, cookie=''):
        """Return LDAP Controls."""
        # set page size for simple paged results
        if not pagesize:
            pagesize = self.pagesize
        # return controls
        return ldapmodule.controls.SimplePagedResultsControl(
            cookie=cookie,
            criticality=True,
            size=pagesize,
        )

    def get_paged_results(self, base, scope, filterstr, attrlist, pagesize=None):
        """Perform a paged search query."""
        items = []

        # define ldap controls for paged results with our pagesize
        lc = self.get_ldap_controls(pagesize)

        # get all pages of results
        cookie = True
        while cookie:

            # get a page of results
            msgid = self.ldapobject.search_ext(
                base=base,
                scope=scope,
                filterstr=filterstr,
                attrlist=attrlist,
                # attrsonly=0,
                serverctrls=[lc],
                # clientctrls=None,
                # timeout=-1,
                # sizelimit=0,
            )

            # get the details of the query result:
            _, rdata, _, serverctrls = self.ldapobject.result3(msgid)

            # add the results data to the items list
            items.extend(rdata)

            # get the cookie for the next request
            cookie = self.set_cookie(lc, serverctrls)

        return items

    def add_entry(self, dn, new):
        """Add an entry in LDAP."""
        ldif = modlist.addModlist(self.convert_entry(new))
        try:
            self.ldapobject.add_s(dn, ldif)
            return True
        except ldapmodule.LDAPError as e:
            print('ERROR adding entry: %s' % (dn))
            print(json.dumps(e.message, indent=2, sort_keys=True))

    def delete_entry(self, dn):
        """Delete an entry in LDAP."""
        try:
            self.ldapobject.delete_s(dn)
            return True
        except ldapmodule.LDAPError as e:
            print('ERROR deleting entry: %s' % (dn))
            print(json.dumps(e.message, indent=2, sort_keys=True))

    def modify_entry(self, dn, current, new):
        """Modify an entry in LDAP."""
        ldif = modlist.modifyModlist(current, new)
        try:
            self.ldapobject.modify_s(dn, ldif)
            return True
        except ldapmodule.LDAPError as e:
            print('ERROR modifying entry: %s' % (dn))
            print(json.dumps(e.message, indent=2, sort_keys=True))

    def create_ou(self, dn):
        """Create a new ou in LDAP."""
        ou = str(dn).split(',')[0].replace('ou=', '')
        record = {
            'objectClass': [
                'top',
                'organizationalUnit',
            ],
            'ou': [ou],
        }
        self.add_entry(dn, record)

    def set_cookie(self, lc, serverctrls):
        """Return the cookie for the next page of results."""
        # get the cookie from the serverctrls
        cookie = serverctrls[0].cookie

        # update the ldap controls object with the new cookie
        lc.cookie = cookie

        # return the cookie
        return cookie



# -*- coding: utf-8 -*-
"""LDAP Update class for ServiceNow."""

from bits.ldap.server import Server


class ServiceNow(Server):
    """ServiceNow LDAP Update class."""

    def get_person(self, person, supervisor_dn):
        """Return a person in the correct format."""
        return person.to_servicenow(supervisor_dn)



# -*- coding: utf-8 -*-
"""LDAP Update Base class for any server type."""

import re


class Server(object):
    """LDAP Server Update class."""

    def __init__(self, update):
        """Initialize a server class instance."""
        self.update = update
        self.verbose = update.verbose

        # ldap class
        self.ldap = None

        # entries from the ldap server
        self.entries = None

        # new entries to be compared
        self.new_entries = None

    #
    # Organization Units
    #
    def add_ous(self):
        """Add any missing ous."""
        # get all ous
        ous = self.get_ous()

        # create missing ous
        for ou in sorted(ous, key=lambda x: len(x)):
            if ou not in self.entries:
                print('    + OU: %s' % (ou))
                # add ou
                self.ldap.create_ou(ou)

        return ous

    def delete_ous(self):
        """Delete any unneeded ous."""
        # get all ous
        ous = self.get_ous(self.entries)
        new_ous = self.get_ous()

        # delete missing ous
        for ou in sorted(ous, key=lambda x: len(x), reverse=True):
            if ou not in new_ous:
                print('    - OU: %s' % (ou))
                # delete ou
                self.ldap.delete_entry(ou)

        return ous

    def get_default_ous(self):
        """Return a list of default ous."""
        return []

    def get_entries_ous(self, entries):
        """Return a list of ous used by all the entries."""
        ous = []
        for dn in entries:
            parent = self.get_ou_parent(dn)
            if re.match('ou=', parent) and parent not in ous:
                ous.append(parent)
        return ous

    def get_entries_to_add(self):
        """Return a list of dns to add."""
        add = []
        for dn in sorted(self.new_entries):
            if dn not in self.entries:
                add.append(dn)
        return add

    def get_entries_to_delete(self):
        """Return a list of dns to delete."""
        delete = []
        for dn in sorted(self.entries):
            if dn not in self.new_entries:
                if not re.match(r'dc=|ou=', dn):
                    delete.append(dn)
        return delete

    def get_entries_to_update(self):
        """Return a list of dns to update."""
        update = []
        for dn in sorted(self.new_entries):
            if dn not in self.entries:
                continue

            # get entries
            entry = self.entries[dn]
            new_entry = self.ldap.convert_entry(self.new_entries[dn])

            # sort mail emails
            # for e in [entry, new_entry]:
            #     for key in ['mail', 'objectClass']:
            #         if key in e and e[key]:
            #             e[key] = sorted(e[key])

            # compare entries
            if entry != new_entry:

                # create update data
                data = {
                    'dn': dn,
                    'entry': entry,
                    'new_entry': new_entry,
                }

                # add data to list
                update.append(data)

        return update

    def get_gnarwl_dn(self, record):
        """Return the proper DN for a gnarwl record."""
        if record.username == 'nagios':
            ou = 'ou=users,%s' % (self.ldap.base_dn)
        else:
            ou = 'ou=people,%s' % (self.ldap.base_dn)
        return 'uid=%s,%s' % (record.username, ou)

    def get_ou_parent(self, ou):
        """Return the parent of an ou."""
        return ','.join(ou.split(',')[1:])

    def get_ou_parents(self, ou):
        """Return the parents of an ou."""
        parents = []

        # return if the ou is not an ou
        if not re.match('ou=', ou):
            return parents

        # get the direct parent of this ou
        parent = self.get_ou_parent(ou)

        # check if the parent is an ou
        if re.match('ou=', parent):
            parents.append(parent)
            # get parents of this ou - commented out recursive call
            # parents.extend(self.get_ou_parents(parent))
        return parents

    def get_ous(self, entries=None):
        """Return a list of all ous."""
        if not entries:
            entries = self.new_entries
        # get default ous from class
        ous = self.get_default_ous()
        # add in ous from each DN in our entries
        ous.extend(self.get_entries_ous(entries))
        # add in any ous that just contain other ous but no records
        ous.extend(self.get_parent_ous(ous))
        # do it once more for good measure (to avoid a recursive call)
        ous.extend(self.get_parent_ous(ous))
        return list(set(ous))

    def get_parent_ous(self, ous):
        """Return a list of ous that are parents of the given ous."""
        parents = []
        for ou in sorted(ous):
            parents.extend(self.get_ou_parents(ou))
        return parents

    def get_person(self, person, supervisor_dn):
        """Return a person in the correct format."""
        return person.to_ldap(supervisor_dn)

    def get_person_dn(self, person):
        """Return the proper DN for a person."""
        return 'uid=%s,ou=people,%s' % (
            person.username,
            self.ldap.base_dn
        )

    def perform_updates(self, server):
        """Peform LDAP updates for a server."""
        if self.verbose:
            print('  Getting entries from LDAP...')
        self.entries = self.ldap.get_entries()
        if self.verbose:
            print('  Retrieved %s entries from LDAP' % (len(self.entries)))

        # display info about entries
        print('  Entries: %s current, %s new' % (
            len(self.entries),
            len(self.new_entries),
        ))

        # add any ous that are missing
        self.add_ous()

        # get entries to add/delete/update
        add = self.get_entries_to_add()
        delete = self.get_entries_to_delete()
        update = self.get_entries_to_update()

        # display info about diff
        print('  Diff: add %s, delete %s, update %s' % (
            len(add),
            len(delete),
            len(update),
        ))

        # perform updates
        self.delete_entries(delete)
        self.add_entries(add)
        self.update_entries(update)

        # delete any ous we no longer need
        self.delete_ous()

    def add_entries(self, add):
        """Add entries to LDAP."""
        if not add:
            return
        for dn in sorted(add):
            print('    + %s' % (dn))
            # add entry
            entry = self.new_entries[dn]
            self.ldap.add_entry(dn, entry)

    def delete_entries(self, delete):
        """Delete entries from LDAP."""
        if not delete:
            return
        for dn in sorted(delete):
            print('    - %s' % (dn))
            # delete entry
            self.ldap.delete_entry(dn)

    def update_entries(self, update):
        """Update entries in LDAP."""
        if not update:
            return

        for item in sorted(update, key=lambda x: x['dn']):
            # get update data
            dn = item['dn']
            entry = item['entry']
            new_entry = item['new_entry']

            # get a list of all keys from both entries
            keys = list(set(sorted(entry) + sorted(new_entry)))

            # create output list
            output = []

            # for key in the account, look for changes
            for key in sorted(keys):

                # get key values
                o = entry.get(key)
                n = new_entry.get(key)

                # order of emails in "mail" doesn't actually matter
                # if key == 'mail':
                #     o = sorted(o)
                #     n = sorted(n)

                # check for changes
                if o != n:
                    output.append('      %s: %s -> %s' % (key, o, n))

            if output:
                # display output
                print('    * %s:' % (dn))
                print('\n'.join(output))

                # update entry
                self.ldap.modify_entry(dn, entry, new_entry)

                # alternate method uses delete, add
                # self.ldap.delete_entry(dn)
                # self.ldap.add_entry(dn, new_entry)

    def prepare_entries(self, ldap):
        """Prepare LDAP entries."""
        self.ldap = ldap
        if self.verbose:
            print('  Preparing entries for LDAP...')
        self.new_entries = self.prepare_people()
        if self.verbose:
            print('  Prepared %s entries for LDAP.' % (len(self.new_entries)))

    def prepare_people(self):
        """Prepare people records for LDAP."""
        people = {}
        for pid in self.update.people_records:
            # get person object
            person = self.update.people_records[pid]
            dn = self.get_person_dn(person)
            # get supervisor object
            supervisor_dn = None
            if person.supervisor_id:
                supervisor = self.update.people_records[person.supervisor_id]
                supervisor_dn = self.get_person_dn(supervisor)
            # get person in ldap format
            people[dn] = self.get_person(person, supervisor_dn)
        return people



# -*- coding: utf-8 -*-
"""Person class file."""


class Person(object):
    """Person class."""

    def __init__(self, update, person):
        """Initialize a person instance."""
        self.person = person
        self.update = update
        self.verbose = update.verbose

        # get people data
        self.get_people_data(person)
        self.get_ad_data()
        self.get_desk_data()
        self.get_google_data()
        self.get_hold_data()
        self.get_nickname_data()
        self.get_phone_data()
        self.get_supervisor_data(person)
        self.get_trusted_tester_data()
        self.get_calculated_fields()

    def get_people_data(self, person):
        """Get data from People."""
        # email address
        email = person.get('email')
        if email:
            self.email = str(email)

        # emplid
        self.emplid = '%s' % (person.get('emplid', ''))

        # end date
        self.end_date = person.get('end_date')

        # first name
        first_name = person.get('first_name')
        if first_name:
            self.first_name = str(first_name)
        else:
            self.first_name = "EMPTY"

        # last name
        last_name = person.get('last_name')
        if last_name:
            self.last_name = str(last_name)
        else:
            self.last_name = "EMPTY"

        # full name
        full_name = person.get('full_name')
        if full_name:
            self.full_name = str(full_name)
        else:
            self.full_name = str(f"{self.first_name} {self.last_name}")

        self.home_institution = person.get('home_institution')
        self.id = person.get('id')

        self.org_unit = person.get('org_unit', '')
        self.start_date = person.get('start_date')
        self.terminated = person.get('terminated')
        self.title = person.get('title')
        self.type = person.get('person_type_name')

        # username
        username = person.get('username')
        if username:
            self.username = username

    def get_ad_data(self):
        """Get data from AD."""
        self.nis_user = None

        if not self.update.ad_users:
            return

        if self.username in self.update.ad_users:
            ad_user = self.update.ad_users[self.username]

            gidNumber = ad_user.get('gidNumber')
            loginShell = ad_user.get('loginShell')
            uidNumber = ad_user.get('uidNumber')
            unixHomeDirectory = ad_user.get('unixHomeDirectory')

            if gidNumber:
                self.gid = gidNumber[0]
            if loginShell:
                self.login_shell = loginShell[0]
            if uidNumber:
                self.uid = uidNumber[0]
            if unixHomeDirectory:
                self.unix_home_directory = unixHomeDirectory[0]

            # check if all nis attributes are set
            if gidNumber and loginShell and uidNumber and unixHomeDirectory:
                self.nis_user = True

    def get_desk_data(self):
        """Get Desk data."""
        self.desk = None

        if self.id in self.update.desks:
            desks = self.update.desks[self.id]

            desk = desks[0]
            self.desk = desk['name']

            building = desk['building']
            self.building = building['code']
            self.city = building['city']
            self.country = 'US'
            self.floor = desk['floor']
            self.state = building['state']
            self.street = '%s, %s-%s' % (
                building['street_address'],
                desk['room']['name'],
                desk['desk']['name'],
            )
            self.zip_code = building['zip']

    def get_hold_data(self):
        """Get Hold data."""
        self.hold = False
        if self.username in self.update.hold:
            self.hold = True

    def get_nickname_data(self):
        """Get Nickname data."""
        self.nicknames = None
        if self.username in self.update.nicknames:
            nicknames = self.update.nicknames[self.username]
            self.nicknames = nicknames['nicknames']

    def get_phone_data(self):
        """Get Phone data."""
        self.phone = None
        if self.id in self.update.phones:
            phones = self.update.phones[self.id]
            phone = phones[0]
            self.phone = phone['phone']

    def get_supervisor_data(self, person):
        """Get Supervisor."""
        self.supervisor_id = None
        supervisor_id = person.get('supervisor_id')
        if supervisor_id and supervisor_id in self.update.people:
            if supervisor_id != self.id:
                self.supervisor_id = supervisor_id

    def get_google_data(self):
        """Get Google Data."""
        self.enterprise = False
        self.url = None

        # set the googe primaryEmail based on the broad email address
        google_email = '%s@broadinstitute.org' % (self.username)

        if google_email in self.update.google_users:
            google_user = self.update.google_users[google_email]
            google_id = google_user['id']

            # check for enterprise license
            if google_id in self.update.enterprise_users:
                self.enterprise = True

            # check for google people record
            if google_id in self.update.google_people:
                person = self.update.google_people[google_id]
                # check for urls
                if 'urls' in person and person['urls']:
                    if not self.url:
                        self.url = str(person['urls'][0]['value'])

    def get_trusted_tester_data(self):
        """Get Trusted Tester data."""
        self.trusted_tester = None
        if self.username in self.update.trusted_testers:
            program = self.update.trusted_testers[self.username]
            self.trusted_tester = program['program']

    def get_calculated_fields(self):
        """Get calculated fields."""
        if self.email:
            self.emails = [self.email]
        if self.org_unit:
            self.organization = self.org_unit.split(' > ')
        # self.future_hire = False

    def to_google(self, supervisor_dn):
        """Return a Person in Google LDAP format."""
        person = self.to_ldap(supervisor_dn)

        # set userPassword
        person['userPassword'] = ['*']

        # set desk info
        if self.desk:
            person['houseIdentifier'] = [self.building.lower()]
            person['postOfficeBox'] = [self.floor]

        # set default gsuite license to Lite:
        gsuite_license = 'Google-Apps-Lite'

        # employees+ get enterprise
        if self.type in ['Core Member', 'Employee', 'Institute Member']:
            gsuite_license = 'Google-Apps-Enterprise'

        # enterprise users also get enterprise
        if self.enterprise:
            # gsuite_license = 'Google-Apps-Enterprise'
            gsuite_license = 'Google-Apps-Unlimited'

        # users on hold get enterprise so they can have vault
        if self.hold:
            gsuite_license = 'Google-Apps-Enterprise'

        # but users who are not on hold and are terminated get Lite.
        elif self.terminated:
            gsuite_license = 'Google-Apps-Lite'

        person['carLicense'] = [gsuite_license]

        return person

    def to_ldap(self, supervisor_dn=None):
        """Return a Person in generic LDAP format."""
        person = {
            # 'businessCategory': [self.org_unit],
            'cn': [self.full_name],
            'co': [self.home_institution],
            'description': [self.email],
            'displayName': [self.full_name],
            'email': [self.email],
            'employeeNumber': [self.emplid],
            'employeeType': [self.type],
            'givenName': [self.first_name],
            'mail': [self.email],
            # 'o': self.organization,
            'objectClass': [
                'top',
                'inetOrgPerson',
                'extensibleObject',
            ],
            'sn': [self.last_name],
            'title': [self.title],
            'uid': [self.username],
        }

        # org_unit info
        if self.org_unit:
            person['businessCategory'] = [self.org_unit]
            person['o'] = self.organization

        # check for desk info
        if self.desk:
            person['buildingName'] = [self.building]
            person['c'] = [self.country]
            person['l'] = [self.city]
            person['postalCode'] = [self.zip_code]
            person['roomNumber'] = [self.desk]
            person['st'] = [self.state]
            person['street'] = [self.street]

        # check for nis info
        if self.nis_user:
            person['uidNumber'] = [self.uid]
            person['gecos'] = [self.full_name]
            person['gidNumber'] = [self.gid]
            person['loginShell'] = [self.login_shell]
            person['homeDirectory'] = [self.unix_home_directory]
            # update objectClass
            person['objectClass'].append('posixAccount')

        # check for phone info:
        if self.phone:
            person['telephoneNumber'] = [self.phone]

        # dn of supervisor
        if supervisor_dn:
            person['manager'] = [supervisor_dn]

        # url:
        if self.url:
            person['info'] = [self.url]

        return person

    def to_servicenow(self, supervisor_dn):
        """Return a Person in Google LDAP format."""
        person = self.to_ldap(supervisor_dn)

        # set shadowExpire
        person['shadowExpire'] = ['0']
        if not self.terminated:
            person['shadowExpire'] = ['1']

        # set userPassword
        if self.nis_user:
            person['userPassword'] = ['{SASL}%s' % (self.username)]

        return person
